""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
from leapp.actors import Actor
from leapp.tags import FortiSOARPhaseTag, IPUWorkflowTag
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.libraries import stdlib
from leapp.libraries.actor.fsr_execute_cmd import fsr_execute_cmd
from leapp.models import FortiSOARFacts
from leapp.exceptions import StopActorExecution, StopActorExecutionError
import os,shutil,subprocess


class FortiSoarPreUpgradeSteps(Actor):
    """
    This actor takes FortiSOAR databasebackup and saves to user provided location
    """

    name = 'fortisoar_pre_upgrade_steps'
    consumes = ()
    produces = (FortiSOARFacts,)
    tags = (IPUWorkflowTag,FortiSOARPhaseTag)
    fortisoar_version = "7.5.0"

    def process(self):

        #If its leapp preupgrade or if its a second run , we will skip execution and return
        if os.environ.get('is_preupgrade') == "true" or os.path.exists('/var/lib/pgsql/14_bkp'):
            api.current_actor().show_message('Identified as a preupgrade call or second execution attempt, not making any changes to the system.')
            api.current_logger().info('Identified as a preupgrade call or second execution attempt, not making any changes to the system.')
            return

        #We will take backup and then clean up FortiSOAR. We will preserve old PGSQL Data folder
        #Collect FortiSOAR information
        product_yum_repo,edition,ver = self._collect_fsr_info()

        #Take FortiSOAR Cofiguration backup and save FortiSOAR info for next actor
        #Function provides directory where backup was takem
        os.environ['is_inplace_upgrade'] = 'true'
        fsr_dir = self._backup_fsr(product_yum_repo,edition,ver)
        
        #Stop FortiSOAR services and rename psql directory
        if edition !="sme":
            self._rename_db_dir()

        #Lets start clean up
        #######################

        #If we see system is SME then only perform certain activity and exit
        if edition == "sme":
            self._clean_up_sme()
            return []

        if edition == "forticloud":
            self._clean_up_forticloud_fsr(fsr_dir)

        self._clean_up_fsr(fsr_dir)

        filepath="/etc/resolv.conf"
        immutable_attr="+i"
        self._change_file_attribute(filepath,immutable_attr)



    #Supporting functions === >
    def _collect_fsr_info(self):
        #Check product repo server
        api.current_actor().show_message('Check FortiSOAR product yum repo server ')
        cmd = ['/bin/cat','/etc/yum/vars/product_yum_server']
        cmd_output = fsr_execute_cmd(cmd)
        product_yum_repo=cmd_output[0]

        #Check FortiSOAR edition
        api.current_actor().show_message('Check FortiSOAR Edition')
        cmd = ['/bin/cat','/etc/cyops-release']
        cmd_output = fsr_execute_cmd(cmd)
        edition = cmd_output[0]
        if "enterprise" in edition:
            edition = "enterprise"
        elif "forticloud" in edition:
            edition = "forticloud"
        else:
            edition = "sme"
        api.current_logger().info('Saved FortiSOAR version and edition details')

        #Check FortiSOAR version
        api.current_actor().show_message('Check FortiSOAR version')
        cmd = ['/bin/cat','/etc/yum/vars/fsr_release_version']
        ver = fsr_execute_cmd(cmd)

        return product_yum_repo,edition,ver[0]

    def _backup_fsr(self,product_yum_repo,edition,ver):
        #Taking FortiSOAR backup

        api.current_actor().show_message('Taking ForiSOAR DB backup. This can take a while...')
        api.current_logger().info('Taking FortiSOAR backup')
        try:
            fobj = open('/etc/leapp/transaction/fsr_backup_directory', 'r')
            fsr_dir = fobj.readline().strip()
            fobj.close()
            api.current_logger().info('Backup directory is "{}"'.format(fsr_dir))
        except OSError as err:
            error = api.current_logger().error('Failed to read `/etc/leapp/transaction/fsr_backup_directory` file: {}'.format(str(err)))
            stdlib.api.current_logger().error(error)
            raise StopActorExecutionError("Failed to read `/etc/leapp/transaction/fsr_backup_directory` file")

        if not os.path.exists('{0}/migrate-fortisoar-{1}.bin'.format(fsr_dir, self.fortisoar_version)):
            error = api.current_logger().error('migrate-fortisoar-{0}.bin not found , exiting : {1}'.format(self.fortisoar_version, str(err)))
            raise StopActorExecutionError(f"migrate-fortisoar-{self.fortisoar_version}.bin not found , exiting")

        os.environ['is_inplace_upgrade'] = 'true'

        bkf_file_name = None
        fsr_cmd = ['/bin/sh', '{0}/migrate-fortisoar-{1}.bin'.format(fsr_dir, self.fortisoar_version), '--export',fsr_dir]
        try:
            output = stdlib.run(fsr_cmd, split=True)['stdout']
            bkf_file_name=next(line for line in output if line.startswith("The Database has been exported into"))
            bkf_file_name=bkf_file_name.split(" ")[-1]
            api.current_logger().info('Backup file saved at "{}"'.format(bkf_file_name))
        except stdlib.CalledProcessError as err:
            error = api.current_logger().error('Failed to run :'+ ' '.join(fsr_cmd) +' {}'.format(str(err)))
            stdlib.api.current_logger().error(error)
            raise StopActorExecutionError("Stopping execution. Data export failed with error: {}".format(str(err)))

        if bkf_file_name :
            # Save FortiSOAR edition , version and backup file path for next actor to use it
            self.produce(FortiSOARFacts(db_filename=bkf_file_name,fsr_version=ver,fsr_edition=edition,product_yum_repo=product_yum_repo,fsr_dir=fsr_dir))
            api.current_logger().info('Saved backupfile info: {0}; FortiSOAR version: {1}; FortiSOAR edition: {2}'.format(bkf_file_name,ver,edition))
            return fsr_dir
        else:
            err_msg = next(line for line in output if line.startswith("Error"))
            raise StopActorExecutionError("Stopping execution. Data export failed with error: {}".format(err_msg))

    def _rename_db_dir(self):
        api.current_logger().info('Stopping all FortiSOAR services')
        api.current_actor().show_message('Stopping all FortiSOAR services')
        cmd = ['csadm','services','--stop']
        fsr_execute_cmd(cmd)

        api.current_actor().show_message('Renaming Postgres data directory')
        cmd = ['/bin/mv','/var/lib/pgsql/14','/var/lib/pgsql/14_bkp']
        fsr_execute_cmd(cmd)

        # In FSR OVA, /var/lib/elasticsearch is a partition and does not get deleted
        # however, added this handling for script installs where it could be only a directory
        api.current_actor().show_message('Renaming ElasticSearch data directory')
        cmd = ['/bin/mv','/var/lib/elasticsearch/nodes','/var/lib/elasticsearch/nodes_bkp']
        fsr_execute_cmd(cmd)

    def _clean_up_sme(self):
        api.current_actor().show_message('Detected Secure Message Exchange')
        api.current_logger().info('Detected Secure Message Exchange')

        api.current_logger().info('Stopping all FortiSOAR services')
        api.current_actor().show_message('Stopping all FortiSOAR services')
        cmd = ['csadm','services','--stop']
        fsr_execute_cmd(cmd)

        packages_to_remove=['cyops-repo-update','cyops-rabbitmq','cyops-common','rabbitmq-server','erlang']
        remove_packages = lambda : [fsr_execute_cmd(['/bin/yum','-y','remove','{}'.format(package)]) for package in packages_to_remove ]
        remove_packages()

        api.current_actor().show_message('Cleaning up /opt')
        dir_to_delete=['cyops','cyops-rabbitmq','cyops-repo-update']
        del_dir = lambda : [fsr_execute_cmd(['/bin/rm','-rf','/opt/{}'.format(dir)]) for dir in dir_to_delete ]
        del_dir()

        api.current_actor().show_message('That\'s all for Secure Message Exchange')
        api.current_logger().info('That\'s all for Secure Message Exchange')

    def _clean_up_forticloud_fsr(self,fsr_dir):
        api.current_logger().info('VM is provisioned on FortiCloud ,hence disabling fsr-init-config and removing cron job')
        api.current_actor().show_message('Disable fsr-init-config services in forticloud VM')
        cmd = ['/bin/systemctl','disable','fsr-init-config']
        fsr_execute_cmd(cmd)

        api.current_actor().show_message('Backup Forticloud flags directory')
        flags_directory = '/opt/cyops/configs/.flags'
        shutil.move(flags_directory,fsr_dir)
        self.produce(FortiSOARFacts(fsr_flags_dir=fsr_dir))

    def _clean_up_fsr(self,fsr_dir):
        api.current_logger().info('Stopping all FortiSOAR services')
        api.current_actor().show_message('Stopping all FortiSOAR services')
        cmd = ['csadm','services','--stop']
        fsr_execute_cmd(cmd)

        api.current_actor().show_message('Cleaning up FortiSOAR yum variables')
        cmd = ['/bin/rm','-rf','remove','/etc/yum/vars/fsr_release_version']
        fsr_execute_cmd(cmd)
        
        api.current_actor().show_message('Make old /opt/cyops-integrations/.env/pip.conf as mutable')
        cmd = ['/bin/chattr','-i','/opt/cyops-integrations/.env/pip.conf']
        fsr_execute_cmd(cmd)
        
        api.current_actor().show_message('Cleaning up old FortiSOAR repo files')
        repo_to_delete = ['fsr-app.repo','fsr-os.repo','fsr-third-party.repo']
        for repo in  repo_to_delete:
            cmd = ['/bin/rm','-f','/etc/yum.repos.d/{}'.format(repo)]
            fsr_execute_cmd(cmd)

        api.current_actor().show_message('Cleaning up rabbitmq files ')
        files_to_delete = ['/etc/rabbitmq/rabbitmq-env.conf','/etc/rabbitmq/rabbitmq.conf', '/etc/rabbitmq/enabled_plugins', '/etc/elasticsearch/jvm.options', '/etc/elasticsearch/elasticsearch.keystore']
        del_file = lambda : [fsr_execute_cmd(['/bin/rm','-rf',file]) for file in files_to_delete ]
        del_file()

        api.current_actor().show_message('Cleaning up FortiSOAR packages and dependencies')
        packages_to_remove = ['php*','perl*','cyops-repo-update','cyops-connector-*','cyops-tomcat','java*','rabbitmq-server','erlang','openssl11-libs','python36-crypto','gd-last', 'elasticsearch','nodejs']
        for package in packages_to_remove:
            cmd = ['/bin/yum','-y','remove',package]
            fsr_execute_cmd(cmd)

        api.current_actor().show_message('Cleaning up /opt')
        dir_to_delete=['cyops','cyops-postgresql','cyops-rabbitmq','cyops-repo-update','cyops-tomcat']
        del_dir = lambda : [fsr_execute_cmd(['/bin/rm','-rf','/opt/{}'.format(dir)]) for dir in dir_to_delete ]
        del_dir()

        api.current_actor().show_message('Cleaning up npm files')
        cmd = ['/bin/rm','-rf','/usr/lib/node_modules/npm/docs']
        fsr_execute_cmd(cmd)

        api.current_actor().show_message('Removing release info')
        cmd = ['/bin/chattr','-i','/etc/cyops-release']
        fsr_execute_cmd(cmd)
        cmd = ['/bin/rm','-f','/etc/cyops-release']

        #Check and remove fm-agent
        api.current_actor().show_message('Checking if ForitMonitor is installed')
        api.current_logger().info('Checking if ForitMonitor is installed')
        rpm_name="fm-agent"
        cmd = '/bin/rpm -q {}'.format(rpm_name)
        try:
            cmd_output = subprocess.call(cmd,shell=True)
            if cmd_output == 0:
                self.produce(FortiSOARFacts(is_fortimonitor_installed=True))
                cmd_remove_rpm= ['/bin/rpm','-e',rpm_name]
                fsr_execute_cmd(cmd_remove_rpm)
        except subprocess.CalledProcessError as e:
            api.current_logger().error('Failed to check if fortimonitor / panopta is installed')
            pass #We don't want to stop execution on non-critical exception

        #Backup yum proxy settings
        api.current_actor().show_message('Backing up yum proxy setting to a file {}/yum_proxy.txt'.format(fsr_dir))
        api.current_logger().info('Backing up yum proxy setting to a file {}/yum_proxy.txt'.format(fsr_dir))
        cmd = '/bin/grep -iF proxy /etc/yum.conf > {}/yum_proxy.txt'.format(fsr_dir)
        try:
            subprocess.call(cmd,shell=True)
        except subprocess.CalledProcessError as e:
            api.current_logger().error('Failed to backup proxy settings')
            pass #We don't want to stop execution on non-critical exception
    
    def _change_file_attribute(self,filepath,attr):
        api.current_actor().show_message('Changing attribute of file {} to {}'.format(filepath,attr))
        api.current_logger().info('Changing attribute of file {} to {}'.format(filepath,attr))
        cmd = ['/bin/chattr',attr,filepath]
        fsr_execute_cmd(cmd)
