""" Copyright start
  Copyright (C) 2008 - 2024 Fortinet Inc.
  All rights reserved.
  FORTINET CONFIDENTIAL & FORTINET PROPRIETARY SOURCE CODE
  Copyright end """
from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import FortiSOARFacts
from leapp.libraries.actor.fsr_execute_cmd import fsr_execute_cmd
from leapp.libraries import stdlib
from leapp.tags import FirstBootPhaseTag, IPUWorkflowTag
import os,time,shutil,subprocess,json


class FortiSoarUpgrade(Actor):
    """
    Execute installer in the freshly booted system, to install FortiSOAR
    """

    name = 'fortisoar_upgrader'
    consumes = (FortiSOARFacts)
    produces = ()
    tags = (IPUWorkflowTag, FirstBootPhaseTag)
    fortisoar_version = "7.5.0"
    ansible_version = "7.4.0"
    FILE_FIRST_BOOT_AFTER_LEAPP_UPGRADE='/opt/cyops/configs/.flags/.first-boot-after-leapp-upgrade'
    FILE_INPLACE_UPGRADE_IN_PROGRESS=f'/home/csadmin/.fortisoar_{fortisoar_version}_inplace_upgrade'
    FILE_LEAPP_UPGRADE_LOG="/var/log/leapp/leapp-upgrade.log"
    FILE_INPLACE_UPGRADE_IN_PROGRESS_MESSAGE=f'/etc/profile.d/fortisoar_{fortisoar_version}_check_inplace_upgrade_status.sh'
    
    def _change_file_attribute(self,filepath,attr):
        api.current_actor().show_message('Changing attribute of file {} to {}'.format(filepath,attr))
        api.current_logger().info('Changing attribute of file {} to {}'.format(filepath,attr))
        cmd = ['/bin/chattr',attr,filepath]
        fsr_execute_cmd(cmd)
    
    def notify_ui_fortisoar_state(self):
        o_fsr_boot_payload={"provision": True,"state": "provisioned","steps" : [], "force_proceed": True}
        f_fsr_state_json = "/opt/cyops-ui/fsr-boot.json"
        f_fsr_state_updater = "/opt/cyops/scripts/state-updater.py"

        subprocess.run([f_fsr_state_updater, "update-state", "--app-name", "provisioning",
                        "--state-file", f_fsr_state_json,
                        "--payload", json.dumps(o_fsr_boot_payload)])

        subprocess.run(["chmod", "644", f_fsr_state_json])
        subprocess.run(["chown", "nginx:nginx", f_fsr_state_json])
    
    def notify_inplace_upgrade_in_progress_on_cli(self):
        # Create Flag to denote the inplace upgrade is in progress.
        
        api.current_actor().show_message('Adding provision to notify user that FortiSOAR Inplace Upgrade is progress.')
        api.current_logger().info('Adding provision to notify user that FortiSOAR Inplace Upgrade is progress.')
        cmd = ['/bin/touch',self.FILE_INPLACE_UPGRADE_IN_PROGRESS]
        fsr_execute_cmd(cmd)
        
        s_show_leapp_upgrade_logs_cmd = f"""
if [ -f "{self.FILE_INPLACE_UPGRADE_IN_PROGRESS}" ]; then 
    echo "=============================================================="
    echo -e "The FortiSOAR {self.fortisoar_version} in-place upgrade is currently in progress. \nPlease monitor the progress in the {self.FILE_LEAPP_UPGRADE_LOG} file and wait for the upgrade process to complete."
    echo "=============================================================="
fi
"""
        with open(self.FILE_INPLACE_UPGRADE_IN_PROGRESS_MESSAGE, 'w') as file:
        # Append the command to the file
            file.write(s_show_leapp_upgrade_logs_cmd + '\n')
        
        # Make the file /etc/profile.d/ readable.
        cmd = ['/bin/chmod','644',self.FILE_INPLACE_UPGRADE_IN_PROGRESS_MESSAGE]
        fsr_execute_cmd(cmd)
            
    def notify_inplace_upgrade_completed_on_cli(self):
        cmd = ['/bin/rm','-f',self.FILE_INPLACE_UPGRADE_IN_PROGRESS]
        fsr_execute_cmd(cmd)
        
        cmd = ['/bin/rm','-f',self.FILE_INPLACE_UPGRADE_IN_PROGRESS_MESSAGE]
        fsr_execute_cmd(cmd)
            
    
    def process(self):
        facts = next(self.consume(FortiSOARFacts), None)
        if not facts:
            return

        bkf_file = facts.db_filename
        fsr_ver = facts.fsr_version
        fsr_edition = facts.fsr_edition
        product_yum_repo=facts.product_yum_repo
        fsr_dir=facts.fsr_dir

        api.current_logger().info("Upgrade configuration - Backup File: {0}, FSR Version: {1}, FSR Edition: {2}, Yum Repository: {3}".format(bkf_file,fsr_ver,fsr_edition,product_yum_repo))

        self.notify_inplace_upgrade_in_progress_on_cli()
        
        #Change PGSQL12 directory name to original one
        if fsr_edition !="sme":
            api.current_actor().show_message('Renaming Postgres data directory to the original location')
            cmd = ['/bin/mv','/var/lib/pgsql/14_bkp','/var/lib/pgsql/14']
            fsr_execute_cmd(cmd)

            # In FSR OVA, /var/lib/elasticsearch is a partition and does not get deleted
            # however, added this handling for script installs where it could be only a directory
            api.current_actor().show_message('Renaming ElasticSearch data directory to the original location')
            cmd = ['/bin/mv','/var/lib/elasticsearch/nodes_bkp','/var/lib/elasticsearch/nodes']
            fsr_execute_cmd(cmd)
            
        #Initialize flag variables.
        flag_ignore_ssl_check = False
        flag_custom_yum_url = False
        s_custom_yum_url = ""
        
        f_custom_yum_url = '{}/custom_yum_url.txt'.format(fsr_dir) 
        api.current_logger().info(f"Checking if the file {f_custom_yum_url} is present.")
        if os.path.exists(f_custom_yum_url) and os.stat(f_custom_yum_url).st_size != 0:
            api.current_logger().info(f"File found at {f_custom_yum_url} indicating custom yum url is exported.")
            flag_custom_yum_url = True
            try: 
                with open(f_custom_yum_url, "r") as file:
                    # Read the content of the file
                    s_custom_yum_url = file.read().strip()
            except Exception as e:
                api.current_logger().warn(f"Could not extract custom_yum_url from {f_custom_yum_url}.")
                api.current_logger().warn(str(e))
            cmd = ['/bin/rm','-f',f_custom_yum_url]
            fsr_execute_cmd(cmd)
                
        f_ignore_ssl_check="{}/ignore_ssl_check.txt".format(fsr_dir) 
        
        api.current_logger().info(f"Checking if the file {f_ignore_ssl_check} is present.")
        if os.path.exists(f_ignore_ssl_check):
            api.current_logger().info(f"File found at {f_ignore_ssl_check} indicating ssl verification is disabled.")
            flag_ignore_ssl_check = True
            cmd = ['/bin/rm','-f',f_ignore_ssl_check]
            fsr_execute_cmd(cmd)
            
        #Remove EL8 packages which are not required
        api.current_actor().show_message('Remove EL8 rpm packages present after leapp upgrade')
        l_conflicting_el8_rpms = ["python36"]
        for s_conflicting_el8_rpm in l_conflicting_el8_rpms :
            cmd = f'/bin/rpm -qa | /bin/grep el8 |  /bin/grep {s_conflicting_el8_rpm} | /bin/xargs rpm -e --nodeps --verbose'
            api.current_logger().info(f"Executing {cmd}.")
            try:
                result = subprocess.run(cmd,shell=True, stdout=subprocess.PIPE,text=True)
                if(result.returncode == 0):
                    api.current_logger().info(f"Removed el8 rpm : {s_conflicting_el8_rpm}.")
                    api.current_logger().info(result.stdout)
                else: 
                    api.current_logger().warning("Error in getting el8 rpms")
            except subprocess.CalledProcessError as e:
                api.current_logger().warning("command to get el8 rpms failed.")
                pass

        #Copy .bash_profile before FSR deployment
        api.current_actor().show_message('Take backup of .bash_profile for csadmin')
        cs_bash_profile= '/home/csadmin/.bash_profile'
        bkf_dir= fsr_dir + "/cs_bash_profile"
        cmd = ['/bin/mkdir','-p',bkf_dir]
        fsr_execute_cmd(cmd)
        if os.path.exists(cs_bash_profile):
            shutil.copy(cs_bash_profile,bkf_dir)

        #Restore yum proxy settings
        file_path = '{}/yum_proxy.txt'.format(fsr_dir)

        yum_conf_file = '/etc/yum.conf' 
        if os.path.exists(file_path) and os.stat(file_path).st_size != 0:
            api.current_actor().show_message('Restoring proxy settings...')
            api.current_logger().info('Restoring proxy settings...')
            cmd = '/bin/cat {0} >> {1}'.format(file_path,yum_conf_file)
            try:
                subprocess.call(cmd,shell=True)
            except subprocess.CalledProcessError as e:
                api.current_logger().error('Failed to restore proxy settings')

        #Required for Cloud VM
        if fsr_edition == "forticloud":
            api.current_actor().show_message('Installing python3-pip')
            p39p_install_cmd = ['/bin/dnf','install','python3-pip','-y']
            fsr_execute_cmd(p39p_install_cmd)
            api.current_logger().info("Installed python3-pip")

        proxy_command = ['grep', '-E', 'https_proxy','/etc/environment']
        https_proxy=None
        try:
           cmd_output = stdlib.run(proxy_command, split=True)['stdout']
           api.current_logger().info(cmd_output)
           if cmd_output and 'https_proxy' in cmd_output[0]:
               api.current_logger().info("Proxy configured in environment file.")
               https_proxy = cmd_output[0].split('https_proxy')[-1].split('=')[-1].strip().strip("\'").strip("\"")
        except Exception as e:
            api.current_logger().warn("Could not extract proxy from environment file.")
            api.current_logger().warn(str(e))

        if https_proxy:
            api.current_logger().info("Found proxy. Adding to environment.")
            os.environ['https_proxy'] = https_proxy
        api.current_actor().show_message('Downloading FortiSOAR install script')
        fsr_install_downld_cmd = ['/bin/wget','-P','{}'.format(fsr_dir),'https://{0}/{1}/install-fortisoar-{2}.bin'.format(product_yum_repo, self.fortisoar_version, self.fortisoar_version)]
        if flag_ignore_ssl_check :
            fsr_install_downld_cmd.append('--no-check-certificate')
        fsr_execute_cmd(fsr_install_downld_cmd)
        api.current_logger().info("Downloaded install script")

        #Set up environment variable for non-interactive installation of FortiSOAR
        os.environ['is_inplace_upgrade'] = 'true'
        if flag_custom_yum_url: 
            os.environ['custom_yum_url'] = s_custom_yum_url
        if fsr_edition == "enterprise" or fsr_edition == "forticloud":
            api.current_actor().show_message('Set up environment variables for the FortiSOAR installer - {}'.format(fsr_edition))
            os.environ['fsr_edition'] = 'enterprise'
            api.current_logger().info("Environment setup done for Enterprise")
        else:
            os.environ['fsr_edition'] = 'secure-message-exchange'
            api.current_logger().info("Environment setup done for SME")

        api.current_actor().show_message('Installing FortiSOAR')
        api.current_logger().info('Installing FortiSOAR')

        fsr_install_cmd = ['/bin/sh','{0}/install-fortisoar-{1}.bin'.format(fsr_dir, self.fortisoar_version)]
        if flag_ignore_ssl_check :
            fsr_install_cmd.append('--ignore-ssl-check')
        fsr_installation_returncode=fsr_execute_cmd(fsr_install_cmd)
        api.current_logger().info("Product installation initiated")
        
        if fsr_installation_returncode == -1 : 
            err_msg = f"Failed to install FortiSOAR {self.fortisoar_version}."
            api.current_logger().error(err_msg)
            raise StopActorExecutionError(err_msg)
        
        if fsr_edition != "sme":
            api.current_actor().show_message('Creating fsr-boot.json file for enterprise edition')
            api.current_logger().info("Creating fsr-boot.json file for enterprise edition")
            self.notify_ui_fortisoar_state()

        if fsr_edition == "forticloud":
            api.current_actor().show_message('Disabling fsr-init-config services')
            api.current_logger().info("Disabling fsr-inti-config service ")
            cmd = ['/bin/systemctl','disable','fsr-init-config']
            fsr_execute_cmd(cmd)

            api.current_actor().show_message('Restoring configuration flags')
            api.current_logger().info("Restoring .flags directory")
            cmd = ['/bin/rm','-rf','/opt/cyops/configs/.flags']
            fsr_execute_cmd(cmd)
            shutil.move(os.path.join(fsr_dir,'.flags'),'/opt/cyops/configs/')
            
            # Create Flag to denote the first boot after inplace upgrade.
            cmd = ['/bin/touch',self.FILE_FIRST_BOOT_AFTER_LEAPP_UPGRADE]
            fsr_execute_cmd(cmd)

            api.current_actor().show_message("Enabling FortiSOAR init service")
            api.current_logger().info("Enabling fsr-init-config")
            cmd = ['/bin/systemctl','enable','fsr-init-config']
            fsr_execute_cmd(cmd)

            api.current_actor().show_message("Starting FortiSOAR init services")
            cmd = ['/bin/systemctl','start','fsr-init-config']
            fsr_execute_cmd(cmd)

        #Wait for  Cloud init to complete
        if fsr_edition == "forticloud":
            api.current_logger().info("Checking if intial setup for FortiCloud is complete")
            cmd = ['/bin/systemctl','status','fsr-init-config']
            cmd_success_msg = "failed with exit code 3"
            wait_time_interval = 60
            while 1:
                api.current_logger().info("Initial setup in progress")
                try:
                    cmd_output = stdlib.run(cmd, split=True)['stdout']['stderr']
                except stdlib.CalledProcessError as err:
                    cmd_output = str(err)
                    pass
                if cmd_success_msg in cmd_output:
                    break
                if (cmd_success_msg not in cmd_output) and itr < 10:
                    time.sleep(wait_time_interval)
                    itr=+1

                    
        # delete the Flag which is no longer required
        cmd = ['/bin/rm','-f',self.FILE_FIRST_BOOT_AFTER_LEAPP_UPGRADE]
        fsr_execute_cmd(cmd)
            
        # Generate encryption key
        api.current_actor().show_message('Generating encryption key')
        api.current_logger().info('Generating encryption key')
        d_config_scripts = "/opt/cyops/scripts"
        f_manage_password = "manage_passwords.py"
        cmd = f"python3 {d_config_scripts}/{f_manage_password} --generate-encryption-key >/dev/null"
        try:
            result = subprocess.run(cmd, shell=True, stdout=subprocess.PIPE, text=True)
            if result.returncode == 0:
                subprocess.run("chmod 644 /opt/cyops/configs/keys/PASSWORD_ENCRYPTION_KEY", shell=True,
                               stdout=subprocess.PIPE, text=True)
            else:
                err_msg = f"Failed to generate encryption key using command {cmd}, exiting"
                api.current_logger().error(err_msg)
                raise StopActorExecutionError(err_msg)
        except subprocess.CalledProcessError as e:
            raise StopActorExecutionError(e)

        # Restore DB
        api.current_logger().info("Restoring the database")
        api.current_actor().show_message('Restoring FortiSOAR version "{}" database and system configuration'.format(fsr_ver))
        restore_cmd = ['/bin/sh','{0}/migrate-fortisoar-{1}.bin'.format(fsr_dir, self.fortisoar_version),'--import',bkf_file]
        restore_cmd_returncode=fsr_execute_cmd(restore_cmd)
        
        if restore_cmd_returncode == -1 : 
            err_msg = f"Failed to restore FortiSOAR FortiSOAR version {fsr_ver} database and system configuration."
            api.current_logger().error(err_msg)
            raise StopActorExecutionError(err_msg)

        # Set password auth, Install ansible for enterprise or forticloud VM  , skip SME
        if fsr_edition != "sme":
            # Restore DB
            api.current_logger().info('Setting the database up for authentication')
            api.current_actor().show_message('Setting the database up for authentication')
            password_cmd = ['sed', '-i', 's/trust/md5/g', '/var/lib/pgsql/16/data/pg_hba.conf']
            fsr_execute_cmd(password_cmd)

            api.current_logger().info("Installing Ansible")
            api.current_actor().show_message('Installing ansible package')
            cmd = ['/bin/sudo','-u','nginx','/opt/cyops-workflow/.env/bin/pip','install','ansible=={0}'.format(self.ansible_version),'--extra-index-url',f'https://{product_yum_repo}/prod/connectors/deps/simple/']
            fsr_execute_cmd(cmd)
            
            # Installing the system widgets after inplace upgrade. This step is run as part of config-vm in case of fresh install.
            api.current_actor().show_message("Installing widgets")
            api.current_logger().info("Installing widgets")
            cmd = ['sudo','-u','nginx','php','/opt/cyops-api/bin/console','app:import:from:file','-f','/opt/cyops-api/src/DataFixtures/Widgets/widgets.json']
            fsr_execute_cmd(cmd)
            
            api.current_actor().show_message("Awaiting for 20 seconds")
            api.current_logger().info("Awaiting for 20 seconds")
            time.sleep(20)

            api.current_logger().info('Running content-hub sync')
            api.current_actor().show_message('Running content-hub sync')
            cmd = ['csadm', 'package', 'content-hub', 'sync', '--force']
            fsr_execute_cmd(cmd)
            
            api.current_logger().info('Restarting services')
            api.current_actor().show_message('Restarting services')
            cmd = ['csadm', 'services', '--restart']
            fsr_execute_cmd(cmd)

        #To prevent eula and vm config from re-executing
        files_to_create=['/home/csadmin/.vm_configured','/home/csadmin/.eula_agreed']
        api.current_actor().show_message('Creating files to prevent EULA and VM config from re-executing')
        for file_name in files_to_create:
            cmd = ['/bin/touch',file_name]
            fsr_execute_cmd(cmd)

        #To prevent config vm exit message and disconnecting shell terminal on first login post OS upgrade
        if os.path.exists(bkf_dir+ "/.bash_profile"):
            shutil.copy(os.path.join(bkf_dir,'.bash_profile'),cs_bash_profile)

        #Updating login banner in /etc/motd

        api.current_logger().info('Update login banner')
        os_version="9" #Use OS version as 9 by default
        cmd = f"sed -n 's/^VERSION_ID=//p' /etc/os-release" #Extract OS version from /etc/os-release
        try:
            result = subprocess.run(cmd,shell=True, stdout=subprocess.PIPE,text=True)
            if(result.returncode == 0):
                os_version = result.stdout.replace('"', '').split("\n")[0]
        except subprocess.CalledProcessError as e:
            pass
        motd_file="/etc/motd"
        os_version_text=f's/^Rocky Linux Version: 8.*/Rocky Linux Version: {os_version}/g'
        cmd = ['/bin/sed','-e',os_version_text,'-i',motd_file]
        fsr_execute_cmd(cmd)

        #Make resolv.conf mutable 
        filepath="/etc/resolv.conf"
        mutable_attr="-i"
        self._change_file_attribute(filepath,mutable_attr)

        #Cleanup /etc/sudoers.d/csadmin file which is removed in vm-config for fresh install.
        f_etc_sudoerd_csadmin="/etc/sudoers.d/csadmin"
        api.current_actor().show_message('Removing /etc/sudoers.d/csadmin')
        api.current_logger().info('Removing /etc/sudoers.d/csadmin')
        cmd = ['/bin/rm','-f',f_etc_sudoerd_csadmin]
        fsr_execute_cmd(cmd)
        
        #Cleanup duplicate entries from /etc/sudoers
        f_etc_sudoers="/etc/sudoers"
        f_etc_sudoers_clean="/etc/sudoers.clean"
        try:
            result = subprocess.run(f"/bin/awk '!a[$0]++' {f_etc_sudoers} > {f_etc_sudoers_clean}",shell=True, stdout=subprocess.PIPE,text=True)
            if(result.returncode == 0):
                subprocess.run(f"cat {f_etc_sudoers_clean} > {f_etc_sudoers}",shell=True, stdout=subprocess.PIPE,text=True)
                subprocess.run(f"rm -f {f_etc_sudoers_clean}",shell=True, stdout=subprocess.PIPE,text=True)
        except subprocess.CalledProcessError as e:
            pass

        #Set crypto policies to DEFAULT
        api.current_actor().show_message('Setting system wide crypto policy back to DEFAULT')
        api.current_logger().info('Setting system wide crypto policy back to DEFAULT')
        cmd = ['update-crypto-policies','--set','DEFAULT']
        fsr_execute_cmd(cmd)
        
        # Set SELinux as enforcing  setenforce Enforcing
        api.current_actor().show_message('Changing SELinux mode to enforcing')
        api.current_logger().info('Changing SELinux mode to enforcing')
        
        # This is needed as leapp adds enforcing=0 entry in /boot/loader/entries/ files which again sets selinux as permissive after boot up.         
        cmd = f"sed -i 's/enforcing=0//g' /boot/loader/entries/*" 
        try:
            result = subprocess.run(cmd,shell=True, stdout=subprocess.PIPE,text=True)
        except subprocess.CalledProcessError as e:
            pass
        
        #Setting selinux mode as Enforcing and persisting the changes.
        cmd = ['/bin/sed', '-i', 's/^SELINUX=permissive/SELINUX=enforcing/g', '/etc/selinux/config']
        fsr_execute_cmd(cmd) 
        cmd = ['setenforce','Enforcing']
        fsr_execute_cmd(cmd) 
        
        self.notify_inplace_upgrade_completed_on_cli()    

        #Adding upgrade complete message in logs and on screen
        #In log
        api.current_logger().info('#######################################################################')
        api.current_logger().info('Process of OS upgrade and FortiSOAR migration is now complete')
        api.current_logger().info('Please check /var/log/leapp/leapp-upgrade.log for any errors if occured')
        api.current_logger().info('#######################################################################')
        if facts.is_fortimonitor_installed:
            api.current_logger().info('#######################################################################')
            api.current_logger().info('Fortimonitor was uninstalled during OS upgrade')
            api.current_logger().info('Please re-install the same once inplace upgrade process completes')
            api.current_logger().info('#######################################################################')


        #On Screen
        api.current_actor().show_message('#######################################################################')
        api.current_actor().show_message('Process of OS upgrade and FortiSOAR migration is now complete')
        api.current_actor().show_message('Please check /var/log/leapp/leapp-upgrade.log for any errors if occured')
        api.current_actor().show_message('#######################################################################')
        if facts.is_fortimonitor_installed:
            api.current_logger().info('#######################################################################')
            api.current_logger().info('Fortimonitor was uninstalled during OS upgrade')
            api.current_logger().info('Please re-install the same once inplace upgrade process completes')
            api.current_logger().info('#######################################################################')
